/****************************************************************************
 *	Copylight(C) 2011-2012 Kanazawa-soft-design,LLC.All Rights Reserved.
 ****************************************************************************/
/*!
 *	@file	wi_rendezvous_port.cpp
 *
 *	@brief	Windows iTRONT[rXR[ (fu).
 *
 *	@author	V 閾
 */
#include <windows.h>
#include <windowsX.h>
#include <stdio.h>
#include "wi_rendezvous_port.h"

/****************************************************************************
 *
 * [J֐`.
 *
 ****************************************************************************/
static WITSKOBJ	*SatisfyRendezvous( WITSKOBJ *tsk, UINT ptn );
static ER		CreateRdv( INT tskid, INT rsize );
static void		DeleteRdv( INT no );

/****************************************************************************
 *
 * [Jϐ`.
 *
 ****************************************************************************/
static INT			RdvSeqNo;		//!< fuԍ̐pJE^.
static WIHDR		*PortList;		//!< fu|[gXg̃|C^.
static WIHDR		*RdvList;		//!< fuXg̃|C^.

/****************************************************************************/
/*!
 *	@brief	fȕ.
 *
 *	@param	Ȃ.
 *
 *	@retval	Ȃ.
 */
void	wi_InitRendPort( void )
{
	RdvSeqNo = 0;
	PortList = NULL;
	RdvList  = NULL;
}

/****************************************************************************/
/*!
 *	@brief	fȕI.
 *
 *	@param	Ȃ.
 *
 *	@retval	Ȃ.
 */
void	wi_QuitRendPort( void )
{
	WIHDR		*next_ptr;
	WIPOROBJ	*port;
	WIRDVOBJ	*rdv;

	//! fu|[gIuWFNg.
	port = (WIPOROBJ *)PortList;
	while( port ){
		next_ptr = port->Hdr.Next;

		//! fu|[gIuWFNg\̂.
		SAFE_RELEASE( port );

		//! ̃fu|[gIuWFNg\̂o.
		port = (WIPOROBJ *)next_ptr;
	}

	//! fuIuWFNg.
	rdv = (WIRDVOBJ *)RdvList;
	while( rdv ){
		next_ptr = rdv->Hdr.Next;

		//! fuIuWFNg\̂.
		SAFE_RELEASE( rdv );

		//! ̃fuIuWFNg\̂o.
		rdv = (WIRDVOBJ *)next_ptr;
	}
}

/****************************************************************************/
/*!
 *	@brief	fu|[g̐.
 *
 *	@param	[in]	id		fu|[gIDԍ.
 *	@param	[in]	obj		fuIuWFNg\̂̃|C^.
 *
 *	@retval	G[R[h.
 */
ER		wi_CreateRendPort( INT id, const WIPOROBJ *obj )
{
	ER			ercd;
	WIPOROBJ	*p;

	//! sȈ̏ꍇ̓G[ɂ.
	if( obj->MaxCalMsg == 0 || obj->MaxRepMsg == 0 ){
		return E_PAR;
	}
	//! fu|[gIuWFNg쐬.
	p = (WIPOROBJ *)wi_CreateObject( id, TMAX_MAXPOR, sizeof(WIPOROBJ), obj, PortList, &ercd );
	if( !p ){
		return ercd;
	}
	//! XgɃfu|[gǉ.
	wi_AddObject( (WIHDR *)p, &PortList );
	return E_OK;
}

/****************************************************************************/
/*!
 *	@brief	fu|[g̍폜.
 *
 *	@param	[in]	id		fu|[gIDԍ.
 *
 *	@retval	G[R[h.
 */
ER		wi_DeleteRendPort( INT id )
{
	ER			ercd;
	WIPOROBJ	*p;

	//! fu|[gID̃IuWFNg擾.
	p = (WIPOROBJ *)wi_FindObject( id, TMAX_MAXPOR, PortList, &ercd );
	if( !p ){
		return ercd;
	}
	//! XgY郉fu|[g菜.
	wi_DelObject( (WIHDR *)p, &PortList );

	//! ҂^XNL[.
	wi_WakeupDelResource( p->CalWaitQue );
	wi_WakeupDelResource( p->AcpWaitQue );

	//! fu|[gIuWFNg\̂.
	SAFE_RELEASE( p );
	return E_OK;
}

/****************************************************************************/
/*!
 *	@brief	fuĂяo.
 *
 *	@param	[in]	id			fu|[gIDԍ.
 *	@param	[in]	ptn			ĂяoĨrbgp^[.
 *	@param	[in]	msg			bZ[W̃AhX.
 *	@param	[in]	csize		ĂяobZ[W̃TCY.
 *	@param	[out]	rsize		bZ[W̃TCYi[̈̃|C^.
 *	@param	[in]	tmout		^CAEgݒ.
 *
 *	@retval	G[R[h.
 */
ER		wi_CallRendPort( INT id, UINT ptn, VP msg, INT csize, INT *rsize, TMO tmout )
{
	ER			ercd;
	WIPOROBJ	*p;
	WITSKOBJ	*acp_tsk,*cal_tsk;

	//! fu|[gID̃IuWFNg擾.
	p = (WIPOROBJ *)wi_FindObject( id, TMAX_MAXPOR, PortList, &ercd );
	if( !p ){
		return ercd;
	}
	//! sȏꍇ̓G[ɂ.
	if( !rsize || p->MaxCalMsg < csize ){
		return E_PAR;
	}
	//! ^XÑ^XNReLXgo.
	cal_tsk = wi_GetTaskObject( TSK_SELF );
	if( !cal_tsk ){
		return E_CTX;
	}
	cal_tsk->Param[0] = (VP)ptn;
	cal_tsk->Param[1] = (VP)msg;
	cal_tsk->Param[2] = (VP)csize;

	//! fut҂Ă^XNɑ΂ăfu邩ǂ𒲂ׂ.
	acp_tsk = SatisfyRendezvous( (WITSKOBJ *)(p->AcpWaitQue), ptn );

	if( !acp_tsk ){
		//! fu܂Ń^XNX[v.
		ercd = wi_TaskWait( id, TTW_CAL, tmout, p->Attribute, &(p->CalWaitQue) );
		if( ercd == E_OK ){
			*rsize = (INT)cal_tsk->Param[2];
		}
	}else{
		//! fuo^.
		ercd = CreateRdv( cal_tsk->Hdr.Id, p->MaxRepMsg );
		if( ercd != E_OK ){
			return ercd;
		}
		//! fuԍt^XNɒʒm.
		acp_tsk->Param[3] = (VP)RdvSeqNo;

		//! fut^XNɃfuĂяobZ[WRs[.
		if( acp_tsk->Param[1] && msg && csize > 0 ){
			memcpy( acp_tsk->Param[1], msg, csize );
			acp_tsk->Param[2] = (VP)csize;
		}else{
			acp_tsk->Param[2] = 0;
		}
		wi_TaskWakeup( acp_tsk->Hdr.Id, TTW_ACP, &(p->AcpWaitQue) );

		//! fuI܂Ń^XNX[v.
		ercd = wi_TaskWait( RdvSeqNo, TTW_RDV, tmout, 0, NULL );
		if( ercd == E_OK ){
			*rsize = (INT)cal_tsk->Param[2];
		}
	}
	return ercd;
}

/****************************************************************************/
/*!
 *	@brief	fu󂯕t.
 *
 *	@param	[in]	id			fu|[gIDԍ.
 *	@param	[in]	ptn			󂯕tĨrbgp^[.
 *	@param	[out]	no			fuԍi[̈̃|C^.
 *	@param	[out]	msg			ĂяobZ[Wi[̈̃|C^.
 *	@param	[out]	size		ĂяobZ[W̃TCYi[̈̃|C^.
 *	@param	[in]	tmout		^CAEgݒ.
 *
 *	@retval	G[R[h.
 */
ER		wi_AcceptRendPort( INT id, UINT ptn, INT *no, VP msg, INT *size, TMO tmout )
{
	ER			ercd;
	WIPOROBJ	*p;
	WITSKOBJ	*cal_tsk,*acp_tsk;

	//! sȏꍇ̓G[ɂ.
	if( !no || !size ){
		return E_PAR;
	}
	//! fu|[gID̃IuWFNg擾.
	p = (WIPOROBJ *)wi_FindObject( id, TMAX_MAXPOR, PortList, &ercd );
	if( !p ){
		return ercd;
	}
	//! ^XÑ^XNReLXgo.
	acp_tsk = wi_GetTaskObject( TSK_SELF );
	if( !acp_tsk ){
		return E_CTX;
	}
	acp_tsk->Param[0] = (VP)ptn;
	acp_tsk->Param[1] = (VP)msg;
	acp_tsk->Param[2] = (VP)0;

	//! fuĂяoĂ^XNɑ΂ăfu邩ǂ𒲂ׂ.
	cal_tsk = SatisfyRendezvous( (WITSKOBJ *)(p->CalWaitQue), ptn );

	if( !cal_tsk ){
		//! fu܂Ń^XNX[v.
		ercd = wi_TaskWait( id, TTW_ACP, tmout, p->Attribute, &(p->AcpWaitQue) );
		if( ercd != E_OK ){
			return ercd;
		}
		*size = (INT)acp_tsk->Param[2];
		*no   = (INT)acp_tsk->Param[3];
	}else{
		//! fuo^.
		ercd = CreateRdv( cal_tsk->Hdr.Id, p->MaxRepMsg );
		if( ercd != E_OK ){
			return ercd;
		}
		//! fuԍ߂lɃZbg.
		*no = RdvSeqNo;

		//! ĂяobZ[WĂяo^XNRs[.
		if( cal_tsk->Param[1] && (INT)cal_tsk->Param[2] > 0 ){
			memcpy( msg, cal_tsk->Param[1], (INT)cal_tsk->Param[2] );
			*size = (INT)cal_tsk->Param[2];
		}else{
			*size = 0;
		}
		//! Ăяo^XN҂s񂩂폜.
		p->CalWaitQue = wi_DelWaitTaskList( cal_tsk->Hdr.Id, p->CalWaitQue );

		//! Ăяo^XÑCxg҂ԂufuI҂vɂ.
		cal_tsk->EventType = TTW_RDV;
		cal_tsk->WaitId    = RdvSeqNo;
		cal_tsk->QueLink   = NULL;

		ercd = E_OK;
	}
	return ercd;
}

/****************************************************************************/
/*!
 *	@brief	fủ
 *
 *	@param	[in]	id		fu|[gIDԍ.
 *	@param	[in]	ptn		󂯕tĨrbgp^[.
 *	@param	[in]	no		fuԍ.
 *	@param	[in]	msg		ĂяobZ[W̃|C^.
 *	@param	[in]	size	ĂяobZ[W̃TCY.
 *
 *	@retval	G[R[h.
 */
ER		wi_FowardRendPort( INT id, UINT ptn, INT no, VP msg, INT size )
{
	ER			ercd;
	WIPOROBJ	*p;
	WITSKOBJ	*acp_tsk,*rdv_tsk;
	WIRDVOBJ	*rdv;

	//! fu|[gID̃IuWFNg擾.
	p = (WIPOROBJ *)wi_FindObject( id, TMAX_MAXPOR, PortList, &ercd );
	if( !p ){
		return ercd;
	}
	//! fuԍ̃IuWFNg擾.
	rdv = (WIRDVOBJ *)wi_FindObject( no, TMAX_MAXRDV, RdvList, &ercd );
	if( !rdv ){
		return E_OBJ;
	}
	//! fuI҂Ă^XÑReLXg擾.
	rdv_tsk = wi_GetTaskObject( rdv->TaskId );
	if( !rdv_tsk ){
		return E_OBJ;
	}
	//! fut҂Ă^XNɑ΂ăfu邩ǂ𒲂ׂ.
	acp_tsk = SatisfyRendezvous( (WITSKOBJ *)(p->AcpWaitQue), ptn );
	if( !acp_tsk ){

		//! fuI҂^XÑCxg҂Ԃufu҂vɂ.
		rdv_tsk->EventType = TTW_CAL;
		rdv_tsk->WaitId    = p->Hdr.Id;

		//! fuĂяo҂sɃfuI҂^XNǉ.
		p->CalWaitQue = wi_AddWaitTaskList( rdv->TaskId, p->Attribute, p->CalWaitQue );
	}else{
		//! fuo^.
		ercd = CreateRdv( rdv_tsk->Hdr.Id, p->MaxRepMsg );
		if( ercd != E_OK ){
			return ercd;
		}
		acp_tsk->Param[2] = 0;
		acp_tsk->Param[3] = (VP)RdvSeqNo;

		//! fut^XNɃfu񑗃bZ[WRs[.
		if( acp_tsk->Param[1] && msg && size > 0 ){
			memcpy( acp_tsk->Param[1], msg, size );
			acp_tsk->Param[2] = (VP)size;
		}
		//! fut҂Ă^XNN.
		wi_TaskWakeup( acp_tsk->Hdr.Id, TTW_ACP, &(p->AcpWaitQue) );
	}
	//! fuIuWFNg폜.
	DeleteRdv( no );
	return E_OK;
}

/****************************************************************************/
/*!
 *	@brief	fȕI.
 *
 *	@param	[in]	rdvno		fuԍ.
 *	@param	[in]	msg			ԓbZ[W̃|C^.
 *	@param	[in]	size		ԓbZ[W̃TCY.
 *
 *	@retval	G[R[h.
 */
ER		wi_ReplayRendPort( INT no, VP msg, INT size )
{
	ER			ercd;
	WIRDVOBJ	*rdv;
	WITSKOBJ	*tsk;

	//! fuԍ̃IuWFNg擾.
	rdv = (WIRDVOBJ *)wi_FindObject( no, TMAX_MAXRDV, RdvList, &ercd );
	if( !rdv ){
		return E_OBJ;
	}
	//! sȏꍇ̓G[ɂ.
	if( rdv->MaxRepMsg < size ){
		return E_PAR;
	}
	//! fuI҂Ă^XÑReLXg擾.
	tsk = wi_GetTaskObject( rdv->TaskId );
	if( !tsk ){
		ercd = E_OBJ;
	}else{
		//! ԓbZ[WfuI҂Ă^XÑbZ[ẄɃRs[.
		ercd = E_OK;
		if( tsk->Param[1] && msg ){
			if( (INT)tsk->Param[2] < size ){
				tsk->Param[2] = 0;
				ercd          = E_PAR;
			}else{
				memcpy( tsk->Param[1], msg, size );
				tsk->Param[2] = (VP)size;
			}
		}else{
			tsk->Param[2] = 0;
		}
		//! fuI҂Ă^XNN.
		wi_TaskWakeup( rdv->TaskId, TTW_RDV, NULL );
	}
	//! fuIuWFNg폜.
	DeleteRdv( no );
	return ercd;
}

/****************************************************************************/
/*!
 *	@brief	fu|[g̏ԎQ.
 *
 *	@param	[in]	id		fu|[gIDԍ.
 *	@param	[out]	sts		fu|[gԂi[\̂̃|C^.
 *
 *	@retval	G[R[h.
 */
ER		wi_RefarenceRendPort( INT id, WIPORSTAT *sts )
{
	ER			ercd;
	WIPOROBJ	*p;

	//! fu|[gID̃IuWFNg擾.
	p = (WIPOROBJ *)wi_FindObject( id, TMAX_MAXPOR, PortList, &ercd );
	if( !p ){
		return ercd;
	}
	//! fu|[g̏Ԃ\̂ɃZbg.
	sts->CalWaitId = wi_GetWaitTaskListFirstId( p->CalWaitQue );
	sts->AcpWaitId = wi_GetWaitTaskListFirstId( p->AcpWaitQue );
	sts->ExtInfo   = p->ExtInfo;
	return E_OK;
}

/****************************************************************************/
/*!
 *	@brief	fuԂ̎Q.
 *
 *	@param	[in]	no		fuԍ.
 *	@param	[out]	sts		fuԂi[\̂̃|C^.
 *
 *	@retval	G[R[h.
 */
ER		wi_RefarenceRendStat( INT no, WIRDVSTAT *sts )
{
	ER			ercd;
	WIRDVOBJ	*p;

	//! fuԍ̃IuWFNg擾.
	p = (WIRDVOBJ *)wi_FindObject( no, TMAX_MAXRDV, RdvList, &ercd );
	if( !p ){
		return ercd;
	}
	//! fȕԂ\̂ɃZbg.
	sts->WaitId = p->TaskId;
	return E_OK;
}

/****************************************************************************/
/*!
 *	@brief	fu邩ǂ𒲂ׂ.
 *
 *	@param	[in]	tsk		҂s̐擪̃^XNIuWFNg̃|C^..
 *	@param	[in]	ptn		Ĩrbgp^[.
 *
 *	@retval	fu^XNIuWFNg̃|C^.
 */
static WITSKOBJ	*SatisfyRendezvous( WITSKOBJ *tsk, UINT ptn )
{
	while( tsk ){
		if( ((UINT)(tsk->Param[0]) & ptn) != 0 ){
			break;
		}
		tsk = tsk->QueLink;
	}
	return tsk;
}

/****************************************************************************/
/*!
 *	@brief	fu̐.
 *
 *	@param	[in]	tskid	fuI҂̃^XNIDԍ.
 *	@param	[in]	rsize	ԓbZ[W̍őTCY.
 *
 *	@retval	G[R[h.
 */
static ER	CreateRdv( INT tskid, INT rsize )
{
	WIRDVOBJ *rdv;

	//! fuԍCNg.
	if( RdvSeqNo >= TMAX_MAXRDV ){
		RdvSeqNo  = 1;
	}else{
		RdvSeqNo++;
	}
	//! fuIuWFNg\̂mۂ.
	rdv = (WIRDVOBJ *)malloc( sizeof(WIRDVOBJ) );
	if( !rdv ){
		return E_NOMEM;
	}
	memset( rdv, 0, sizeof(WIRDVOBJ) );
	rdv->Hdr.Id    = RdvSeqNo;
	rdv->TaskId    = tskid;
	rdv->MaxRepMsg = rsize;

	//! Xgɐfuǉ.
	wi_AddObject( (WIHDR *)rdv, &RdvList );
	return E_OK;
}

/****************************************************************************/
/*!
 *	@brief	fu̍폜.
 *
 *	@param	[in]	no		fuԍ.
 *
 *	@retval	Ȃ.
 */
static void	DeleteRdv( INT no )
{
	ER			ercd;
	WIRDVOBJ	*p;

	//! fu|[gID̃IuWFNg擾.
	p = (WIRDVOBJ *)wi_FindObject( no, TMAX_MAXRDV, RdvList, &ercd );
	if( p ){
		//! XgY郉fu|[g菜.
		wi_DelObject( (WIHDR *)p, &RdvList );

		//! fuIuWFNg\̂.
		SAFE_RELEASE( p );
	}
}

/****************************************************************************/
/*!
 *	@brief	폜^XN҂s񂩂폜.
 *
 *	@param	[in]	id		폜^XNIDԍ.
 *
 *	@retval	Ȃ.
 */
void	wi_AllReleaseRendPort( INT id )
{
	WIPOROBJ *p;

	for( p = (WIPOROBJ *)PortList; p; p = (WIPOROBJ *)p->Hdr.Next ){

		//! 폜^XN҂s񂩂폜.
		p->CalWaitQue = wi_DelWaitTaskList( id, p->CalWaitQue );
		p->AcpWaitQue = wi_DelWaitTaskList( id, p->AcpWaitQue );
	}
}

/* End of file */
